class BoardStack:

    def __init__(self):
        self.past_board_states = []

    def push(self, board):
        self.past_board_states.append(str(board))
        if len(self.past_board_states) > 2:
            del self.past_board_states[0]

    def breaks_simple_ko_rule(self, current_board):
        if len(self.past_board_states) < 2:
            return False
        return str(current_board) == self.past_board_states[-2]

    def reset(self):
        self.past_board_states = []

    def remove_last(self):
        if len(self.past_board_states):
            del self.past_board_states[-1]



from baduk.action.load_sgf import LoadSGFAction
from baduk.action.start_game import StartGameAction
from baduk.ui.action.action_collection import ActionCollection
from baduk.ui.action.action_item import ActionItem


class MenuActionDialog:

    @staticmethod
    def main_menu():
        return ActionCollection(
            ActionItem('p', 'Play Game', StartGameAction()),
            ActionItem('l', 'Load SGF', LoadSGFAction()),
            ActionItem('q', 'Quit Game'),
        )



from ansi_colours import AnsiColours as Colour
from terminal_table.table import Table

from baduk.ui.style import Style
from baduk.ui.user_input import UserInput
from baduk.ui.view import View


class Pagination:

    def __init__(self, arr, headers):
        self.arr = arr
        self.headers = headers

    def __call__(self, limit=5, find=None, find_by_id=None):
        self.limit = limit
        self.total = len(self.arr)
        self.page = 1
        return self.menu()

    def menu(self):
        result = False
        while not result:
            View().render(text=self.paginated_table())
            View().render(text=self.pagination_text())
            user_input = UserInput.get_input('\nEnter a %s, use %s and %s to navigate, or %s to go back: ' % (
                Colour.green("'number'"),
                Colour.green("<"),
                Colour.green(">"),
                Colour.green("b")
            ))
            if user_input == 'b':
                break
            if self.should_change_page(user_input):
                self.change_page(user_input)
                continue
            if user_input not in [str(x) for x in range(1, len(self.arr) + 1)]:
                continue
            result = self.arr[int(user_input) - 1]
            View().render(text="An item with that key was not found. Press any key to continue.")
        return result

    def paginated_table(self):
        arr_start = (self.page * self.limit) - self.limit
        arr_end = (self.page * self.limit)
        return Table.create(
            self.arr[arr_start:arr_end],
            self.headers
        )

    def pagination_text(self):
        text = ""
        if self.page > 1:
            text += "< previous | "
        text += "page %d" % self.page
        if self.page * self.limit < self.total:
            text += " | next >"
        output = Style.create_underline(text)
        output += "\n%s\n" % text
        output += Style.create_underline(text)

        return output

    def should_change_page(self, user_input):
        return user_input is '<' or user_input is '>'

    def change_page(self, user_input):
        if self.page is not 1 and user_input is '<':
            self.page -= 1
        elif self.page * self.limit < self.total and user_input is '>':
            self.page += 1



from text_template import TextTemplate

from baduk.constants.resource_path import TEMPLATE_PATH


class View:

    def __init__(self, template="blank.txt"):
        self.template_path = TEMPLATE_PATH + template

    def render(self, **kwargs):
        print(TextTemplate.render(self.template_path, **kwargs))



from baduk.constants.constants import ALPHA_KEY
from baduk.enums.enums import Stone
from baduk.game.board_stack import BoardStack
from baduk.game.point import Point
from baduk.stones.group_of_stones_collection import GroupOfStonesCollection
from baduk.stones.stone_link import StoneLink
from baduk.validation.move_validation import MoveValidation


class Board:
    def __init__(self, size: tuple):
        self.size = size
        self.board = self.make_board()
        self.group_collection = GroupOfStonesCollection()
        self.valid_move = False
        self.board_stack = BoardStack()

    def __repr__(self):
        return '\n'.join([row for row in map(lambda x: ' '.join(map(lambda y: y.stone.value, x)), self.board)])

    def get_size(self):
        return self.size

    def get_point_stone_link(self, x, y):
        return self.board[y][x]

    def is_valid_move(self):
        return self.valid_move

    def draw(self):
        alpha_row = ' '.join([str(ALPHA_KEY[i]) for i in range(self.size[0])])
        rows = '\n'.join(['%02s %s' % (i + 1, row) for i, row in
                          enumerate(map(lambda x: ' '.join(map(lambda y: y.stone.value, x)), self.board))])
        print('   %s\n%s' % (alpha_row, rows))

    def in_grid(self, point):
        return 0 <= point['x'] < self.size[0] and 0 <= point['y'] < self.size[1]

    def place_stone(self, point: Point, stone: Stone):
        MoveValidation.check_point_is_in_bounds(point, self.size)
        MoveValidation.check_stone_link_is_empty(self.get_point_stone_link(point.x, point.y))
        stone_link = self.get_point_stone_link(point.x, point.y)
        stone_link.set_stone(stone)
        self.group_collection.add_stone_link(stone_link, self)
        self.board_stack.push(self)


    def reset(self):
        self.group_collection.reset()
        self.board_stack.reset()
        self.board = self.make_board()

    def make_board(self):
        return [[StoneLink(Point(x, y), Stone.NONE) for x in range(self.size[0])] for y in range(self.size[1])]

    def rollback(self):
        self.board_stack.remove_last()
        self.group_collection.rollback()
        self.group_collection.update_groups_of_stones(self)



from random import randint, random

from baduk.constants.constants import HANDICAP_9_X_9, HANDICAP_13_X_13, HANDICAP_19_X_19
from baduk.enums.enums import Stone
from baduk.exception.exceptions import ValidationError
from baduk.game.board import Board
from baduk.game.move_log import MoveLog
from baduk.game.player import Player
from baduk.game.point import Point
from baduk.game.turn import Turn
from baduk.service.move_from_sgf import MovesFromSGF


class Game:

    def __init__(self, size: int, height: int = None):
        if size > 25 or height is not None and height > 25:
            raise ValidationError('Board must be less than 25 x 25')
        if height is not None:
            size = (size, height)
        else:
            size = (size, size)
        self.size = size
        self.height = height
        self._board = Board(size)
        self.players = (Player('Black', Stone.BLACK), Player('White', Stone.WHITE))
        self._turn = Turn()
        self.move_log = MoveLog()
        self.passes = 0

    def handicap(self, stones):
        if self._turn.get_turn() > 1:
            raise ValidationError('Handicap stones cannot be placed after first move has been taken')
        if self.size[0] == 19 and self.size[1] == 19:
            self._place_handicap_stones(stones, HANDICAP_19_X_19)
        elif self.size[0] == 13 and self.size[1] == 13:
            self._place_handicap_stones(stones, HANDICAP_13_X_13)
        elif self.size[0] == 9 and self.size[1] == 9:
            self._place_handicap_stones(stones, HANDICAP_9_X_9)
        else:
            raise ValidationError('Handicap stones can only be placed on 19x19, 13x13 and 9x9 boards')

    def _place_handicap_stones(self, stones, handicap_stones):
        current_player = self.get_current_player()
        for i in range(stones):
            self._make_move(handicap_stones[i], current_player)
        self._turn.next_turn()

    def board(self):
        self._board.draw()

    def move(self, *moves):
        for move in moves:
            if move == 'pass':
                self.pass_turn()
                return
            self.passes = 0
            current_player = self.get_current_player()
            self._make_move(move, current_player)
            current_player.update_killed_stone_count(self._board.group_collection.get_dead_stones_count())
            self.move_log.add(move)
            self._turn.next_turn()

    def _make_move(self, coordinate, current_player):
        point = Point(coordinate=coordinate)
        stone = current_player.get_stone()
        self._board.place_stone(point, stone)

    def rollback(self, number_of_turns: int):
        for _ in range(number_of_turns):
            if self.move_log.count == 0:
                raise ValidationError("Too many rollbacks")
            if self.move_log.pop() != 'pass':
                self._board.rollback()
            self._turn.rollback_turn()

    def get_current_player(self):
        return self.players[self._turn.current_player() - 1]

    def get_position(self, coordinate):
        point = Point(coordinate=coordinate)
        return self._board.get_point_stone_link(point.x, point.y).stone.value

    def pass_turn(self):
        self.passes += 1
        self.move_log.add('pass')
        if self.passes == 2:
            self.end_game()
        self._turn.next_turn()

    def turn(self):
        return 'black' if self._turn.get_turn() % 2 else 'white'

    def reset(self):
        self.passes = 0
        self._board.reset()
        self._turn.reset()
        self.move_log.reset()

    def end_game(self):
        print('End Game.')

    def replay_sgf(self, sgf):
        moves = MovesFromSGF(sgf).get_as_korschelt()
        self.move(*moves)
        return len(moves)


if __name__ == '__main__':
    game = Game(19)

    from os import walk

    sgfs = []
    for (dir_path, dir_names, file_names) in walk('../../sgf'):
        sgfs.extend(file_names)
        break

    for sgf in sgfs:
        moves = MovesFromSGF('../../sgf/' + sgf).get_as_korschelt()
        index = 0
        while index < len(moves):
            try:
                game.move(moves[index])
            except:
                index += 1
                continue
            if random() > 0.99 and index > 10:
                rollback = int(randint(3, 10))
                game.rollback(rollback)
                index -= (rollback - 1)
                continue
            index += 1
        game.board()
        game.reset()
    game = Game(19)
    game.board()
    game.reset()
    game = Game(9)
    board_point = Point(coordinate='1A')
    assert board_point.x == 0 and board_point.y == 0
    board_point = Point(coordinate='9J')
    assert board_point.x == 8 and board_point.y == 8
    moves = ["4D", "3D", "4H", "5D", "3H", "4C", "5B", "4E"]
    game.move(*moves)
    game.board()
    assert game.get_position('4D') == '.'
    game.reset()
    moves = ["6D", "7E", "6E", "6F", "4D", "5E", "5D", "7D",
             "5C", "6C", "7H", "3D", "4E", "4F", "3E", "2E",
             "3F", "3G", "2F", "1F", "2G", "2H", "1G", "1H",
             "4C", "3C", "6H", "4B", "5H", "5B"]
    captured = ["6D", "6E", "4D", "5D", "5C", "4E", "3E", "3F", "2F", "2G", "1G", "4C"]
    game.move(*moves)
    for capture in captured:
        assert game.get_position(capture) == '.'
    game.board()
    game.reset()
    moves = ["9A", "8A", "8B", "9B"]
    game.move(*moves)
    assert game.get_position('9A') == '.'
    game.board()
    game.reset()
    moves = ["5D", "5E", "4E", "6E", "7D", "4F", "7E", "3E", "5F", "4D",
             "6F", "6D", "6C", "7F", "4E", "5E"]
    captured = ["4E", "6D", "6E"]
    game.move(*moves)
    for capture in captured:
        assert game.get_position(capture) == '.'
    game.board()
    game.reset()
    small_game = Game(5)
    moves = ["5A", "1E", "5B", "2D", "5C", "2C", "3A",
             "1C", "2A", "3D", "2B", "3E", "4D", "4B",
             "4E", "4A", "3C", "3B", "1A", "4C", "3C"]
    captured = ["4A", "4B", "4C", "3B"]
    small_game.move(*moves)
    for capture in captured:
        assert small_game.get_position(capture) == '.'
    small_game.reset()
    moves = ["2A", "5A", "1B", "5B", "3B", "5C", "2C", "2B", "5D"]
    for move in moves:
        try:
            small_game.move(move)
        except:
            continue
    assert small_game.get_position('2B') == '.'
    assert small_game.get_position('5D') == 'o'
    small_game.reset()
    moves = ["2A", "1C", "1B", "2D", "3B", "3C", "2C", "2B", "2C", "5A"]
    for move in moves:
        try:
            small_game.move(move)
        except:
            continue
    small_game.board()
    assert small_game.get_position('2B') == 'o'
    assert small_game.get_position('2C') == '.'
    assert small_game.get_position('5A') == 'x'
    small_game.reset()
    moves = ["5C", "5B", "4D", "4A", "3C", "3B", "2D", "2C", "4B", "4C", "4B", "2B"]
    for move in moves:
        try:
            small_game.move(move)
        except:
            continue
    small_game.board()
    assert small_game.get_position("2B") == 'x'
    assert small_game.get_position("4B") == '.'
    small_game.reset()
    moves = ["1A", "2A"]
    for move in moves:
        try:
            small_game.move(move)
        except:
            continue
    small_game.board()
    small_game.rollback(1)
    small_game.board()
    assert small_game.get_position("1A") == 'x'
    assert small_game.get_position("2A") == '.'
    small_game.reset()
    moves = ["1A", "2A", "1E", "1B"]
    for move in moves:
        try:
            small_game.move(move)
        except:
            continue
    small_game.board()
    assert small_game.get_position("1A") == '.'
    assert small_game.get_position("1B") == 'o'
    small_game.rollback(1)
    small_game.board()
    assert small_game.get_position("1A") == 'x'
    assert small_game.get_position("1B") == '.'
    small_game.reset()
    moves = ["1A", "2A", "1E", "1C"]
    for move in moves:
        try:
            small_game.move(move)
        except:
            continue
    small_game.board()
    small_game.pass_turn()
    moves = ["4B", "4C"]
    for move in moves:
        try:
            small_game.move(move)
        except:
            continue
    small_game.board()
    assert small_game.get_position("1C") == 'o'
    assert small_game.get_position("4B") == 'o'
    assert small_game.get_position("4C") == 'x'
    small_game.rollback(3)
    small_game.board()
    assert small_game.get_position("1C") == 'o'
    assert small_game.get_position("4B") == '.'
    assert small_game.get_position("4C") == '.'



from baduk.command.command_types import Command, UndoableCommand


class ChainOfCommands:
    stack = []

    def execute_command(self, command: Command):
        result = command.execute()
        if isinstance(command, UndoableCommand):
            self.stack.append(command)
        return result

    def undo(self):
        if len(self.stack):
            command = self.stack.pop()
            command.undo()

    def reset(self):
        self.stack = []



from baduk.ui.action.action_collection import ActionCollection
from baduk.ui.action.action_item import ActionItem


class BoardSetUpDialog:

    @staticmethod
    def select_board_size():
        return ActionCollection(
            ActionItem(range(5, 26), 'Select Board Size'),
        )

    @staticmethod
    def set_handicap():
        return ActionCollection(
            ActionItem(range(0, 10), 'Set Handicap'),
        )



from baduk.constants.constants import ALPHA_KEY
from baduk.ui.action.action_collection import ActionCollection
from baduk.ui.action.action_item import ActionItem


class PlayerActionDialog:

    def __init__(self, size):
        self.keys = ["%s%s" % (i + 1, ALPHA_KEY[j]) for i in range(size) for j in range(size)]

    def player_turn(self):
        return ActionCollection(
            ActionItem(self.keys, 'Enter Move'),
            ActionItem('p', 'Pass'),
            ActionItem('u', 'Undo'),
        )



from baduk.enums.enums import Stone
from baduk.exception.exceptions import ValidationError


class MoveValidation:
    @staticmethod
    def check_stone_link_is_empty(stone_link):
        if stone_link.stone != Stone.NONE:
            raise ValidationError('Stone cannot be placed on another stone')

    @staticmethod
    def check_point_is_in_bounds(point, size):
        if point.x < 0 or point.x > size[0] - 1 or point.y > size[1] - 1 or point.y < 0:
            raise ValidationError('Stone placed out of bounds')

    @staticmethod
    def check_self_capture(group, has_dead_stones, chain_of_commands):
        if not has_dead_stones and group.get_liberties() == 0:
            chain_of_commands.undo()
            raise ValidationError("Self capture is illegal")

    @staticmethod
    def breaks_simple_ko_rule(group_of_stones_collection, dead_stones, board):
        if group_of_stones_collection.dead_stone_count == 1 and board.board_stack.breaks_simple_ko_rule(board):
            group_of_stones_collection.chain_of_commands.undo()
            group_of_stones_collection.chain_of_commands.undo()
            group_of_stones_collection.update_groups_of_stones(board, dead_stones)
            raise ValidationError('Breaks simple Ko Rule')



from baduk.command.command_types import UndoableCommand


class RemoveDeadStones(UndoableCommand):

    def __init__(self, groups: set, dead_stone_groups: list):
        self.groups = groups
        self.group = None
        self.dead_stone_groups = dead_stone_groups

    def execute(self):
        if len(self.dead_stone_groups) > 0:
            for dead_stone_group in self.dead_stone_groups:
                dead_stone_group.capture_stones()
                self.groups.remove(dead_stone_group)

    def undo(self):
        for dead_stone_group in self.dead_stone_groups:
            dead_stone_group.replace_captured_stones()
            self.groups.add(dead_stone_group)



from baduk.constants.constants import ALPHA_KEY
from baduk.exception.exceptions import ValidationError


class Point:
    coordinate = None
    x = None
    y = None

    def __init__(self, x=None, y=None, coordinate=None):
        if coordinate is not None and len(coordinate) > 3:
            raise ValidationError('Coordinate format is invalid')
        if coordinate:
            self.parse_coordinate(coordinate)
        elif x is not None and y is not None:
            self.parse_x_y(x, y)
        else:
            raise ValidationError('Needs to be passed a coordinate or x and y')

    def __repr__(self):
        return self.coordinate

    def parse_coordinate(self, coordinate):
        self.coordinate = coordinate
        alpha = coordinate[-1:]
        self.x = int(ALPHA_KEY.index(alpha))
        self.y = int(coordinate[:-1]) - 1

    def parse_x_y(self, x, y):
        self.coordinate = '%d%s' % (y + 1, ALPHA_KEY[x])
        self.x = x
        self.y = y

    def adjacent_points(self):
        return [
            {"x": self.x + 1, "y": self.y},
            {"x": self.x - 1, "y": self.y},
            {"x": self.x, "y": self.y + 1},
            {"x": self.x, "y": self.y - 1}
        ]

    def map_adjacent_links(self, callback, board):
        for adjacent_point in self.adjacent_points():
            if board.in_grid(adjacent_point):
                yield callback(board.get_point_stone_link(**adjacent_point))

    def filter_adjacent_links(self, callback, board):
        for adjacent_point in self.adjacent_points():
            if board.in_grid(adjacent_point):
                link = board.get_point_stone_link(**adjacent_point)
                if callback(link):
                    yield link



from abc import ABCMeta, abstractmethod


class Action(metaclass=ABCMeta):

    def __call__(self, input_value):
        self.execute(input_value)

    @abstractmethod
    def execute(self, input_value):
        pass



from baduk.command.command_types import UndoableCommand
from baduk.stones.group_of_stones import GroupOfStones
from baduk.stones.stone_link import StoneLink


class CreateNewGroupOfStones(UndoableCommand):

    def __init__(self, groups: set, stone_link: StoneLink):
        self.stone_link = stone_link
        self.groups = groups
        self.group = None

    # Todo: Command + Undo: remove the new group
    def execute(self, **kwargs):
        self.group = GroupOfStones(self.stone_link)
        self.groups.add(self.group)
        return self.group

    def undo(self):
        self.group.remove_stone_from_board(self.stone_link)
        self.groups.remove(self.group)



class ActionCollection:

    def __init__(self, *actions):
        self.actions = []
        self.count = 0
        for action in actions:
            self.count += 1
            self.actions.append(action)

    def add_action(self, action):
        self.count += 1
        self.actions.insert(-1, action)

    def add_actions(self, *actions):
        for action in reversed(actions):
            self.count += 1
            self.actions.insert(-1, action)



from abc import ABCMeta, abstractmethod


class Command(metaclass=ABCMeta):

    @abstractmethod
    def execute(self):
        pass


class UndoableCommand(Command, metaclass=ABCMeta):

    @abstractmethod
    def undo(self):
        pass



from baduk.dialog.board_set_up_dialog import BoardSetUpDialog
from baduk.dialog.player_action_dialog import PlayerActionDialog
from baduk.exception.exceptions import ValidationError
from baduk.game.game import Game
from baduk.ui.action.action import Action
from baduk.ui.user_menu import UserMenu
from baduk.ui.view import View


class StartGameAction(Action):

    def execute(self, input_value):
        size = int(UserMenu(BoardSetUpDialog.select_board_size(), "Select Board Size")())
        game = Game(size)
        if size in [9, 13, 19]:
            handicap = UserMenu(BoardSetUpDialog.set_handicap(), "Set Handicap")()
            game.handicap(int(handicap))
        game.board()
        player_dialog = PlayerActionDialog(size)
        passes = 0
        while passes < 2:
            move = UserMenu(player_dialog.player_turn(), "Place Stone")()
            if move == 'p':
                passes += 1
                game.board()
            elif move == 'u':
                game.rollback(1)
                game.board()
            else:
                passes = 0
                try:
                    game.move(move)
                    game.board()
                except ValidationError as error:
                    game.board()
                    View("error.txt").render(error=error)



from baduk.dialog.menu_action_dialog import MenuActionDialog
from baduk.ui.user_menu import UserMenu


class Main:
    def __init__(self):
        while True:
            if UserMenu(MenuActionDialog.main_menu())() == 'q':
                break


if __name__ == '__main__':
    Main()



class MoveLog:

    def __init__(self):
        self.moves = []
        self.count = 0

    def add(self, move):
        self.moves.append(move)
        self.count += 1

    def pop(self):
        self.count -= 1
        return self.moves.pop()

    def count(self):
        return self.count

    def reset(self):
        self.moves = []
        self.count = 0



from baduk.command.command_types import UndoableCommand
from baduk.stones.stone_link import StoneLink


class AddToAdjacentGroupOfStones(UndoableCommand):

    def __init__(self, groups: set, stone_link: StoneLink, adjacent_stones: list):
        self.stone_link = stone_link
        self.adjacent_stones = adjacent_stones
        self.groups = groups
        self.group = None

    # Todo: Command + Undo: remove stone_link from group
    def execute(self, **kwargs):
        adjacent_stone = self.adjacent_stones[0]
        self.group = adjacent_stone.group
        self.group.add_link(self.stone_link)

        return self.group

    def undo(self):
        self.group.remove_stone_from_board(self.stone_link)



from ansi_colours import AnsiColours as Colour

from baduk.ui.action.action_collection import ActionCollection
from baduk.ui.user_input import UserInput
from baduk.ui.view import View


class UserMenu:

    def __init__(self, action_collection: ActionCollection, title="Menu"):
        self.action_collection = action_collection
        self.title = title

    def __call__(self):
        user_input = None
        keys = {}
        View("menu.txt").render(title=self.title, menu=self._prepare_menu(keys))
        while user_input not in keys:
            user_input = UserInput.get_input('Select an option: ')
        return keys[user_input](user_input)

    def _prepare_menu(self, keys):
        menu = ''
        for action in self.action_collection.actions:
            if isinstance(action.key, range):
                menu += "%s: %d-%d\n" % (str(action), action.key[0], action.key[-1])
                keys.update({str(key): action.execute for key in action.key})
            elif isinstance(action.key, list):
                menu += "%s: %s-%s\n" % (str(action), action.key[0], action.key[-1])
                keys.update({str(key): action.execute for key in action.key})
            else:
                menu += "%s: %s\n" % (str(action), Colour.green(action.key))
                keys[str(action.key)] = action.execute
        return menu



from baduk.enums.enums import Stone
from baduk.stones.stone_link import StoneLink


class GroupOfStones:
    def __init__(self, link: StoneLink):
        self.stone = link.stone
        self.links = set()
        self.liberties = set()
        link.set_group(self)
        self.first_link = link
        self.links.add(link)

    #
    # def __repr__(self):
    #     return 'Key: %s | Liberties %d' % (
    #         (', '.join([str(link) for link in self.links])), len(self.liberties))

    def update(self, board):
        self.links.clear()
        self.links.add(self.first_link)
        self.liberties.clear()
        self.first_link.find_adjacent(self.links, self.liberties, board)

    def add_link(self, link: StoneLink):
        link.set_group(self)
        self.links.add(link)

    def get_liberties(self):
        return len(self.liberties)

    def get_links(self):
        return self.links

    def merge(self, group):
        self.links = group.links | self.links
        self.liberties = group.liberties | self.liberties

    def set_links_group(self):
        for link in self.links:
            link.set_group(self)

    def count_stones(self):
        return len(self.links)

    def remove_stone_from_board(self, stone_link):
        stone_link.set_stone(Stone.NONE)
        stone_link.set_group(None)
        self.links.remove(stone_link)

    def capture_stones(self):
        for link in self.links:
            link.set_stone(Stone.NONE)
            link.set_group(None)

    def replace_captured_stones(self):
        for link in self.links:
            link.set_stone(self.stone)
            link.set_group(self)



from baduk.enums.enums import Stone


class Player:
    def __init__(self, name, stone: Stone):
        self.stone = stone
        self.name = name
        self.killed_stones = 0

    def get_stone(self):
        return self.stone

    def update_killed_stone_count(self, killed_stone_count):
        self.killed_stones += killed_stone_count

    def __str__(self):
        return self.name



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['baduk',
 'baduk.action',
 'baduk.command',
 'baduk.constants',
 'baduk.dialog',
 'baduk.enums',
 'baduk.exception',
 'baduk.game',
 'baduk.service',
 'baduk.stones',
 'baduk.ui',
 'baduk.ui.action',
 'baduk.validation']

package_data = \
{'': ['*'], 'baduk': ['templates/*']}

install_requires = \
['ansi-colours==0.2.6', 'terminal-table==0.3.1', 'text-template==0.1.4']

setup_kwargs = {
    'name': 'baduk',
    'version': '1.0.1',
    'description': 'Python Baduk Game',
    'long_description': None,
    'author': 'Sarcoma',
    'author_email': 'sean@orderandchaoscreative.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from baduk.command.command_types import UndoableCommand
from baduk.stones.group_of_stones import GroupOfStones
from baduk.stones.stone_link import StoneLink


class MergeWithAdjacentGroupsOfStones(UndoableCommand):

    def __init__(self, groups: set, stone_link: StoneLink, adjacent_stones: list):
        self.stone_link = stone_link
        self.adjacent_stones = adjacent_stones
        self.groups = groups
        self.group = None
        self.old_groups = set()

    # Todo: Command + Undo: remove the new group, add the old groups, update the links group
    def execute(self, **kwargs):
        self.group = GroupOfStones(self.stone_link)
        for adjacent_stone in self.adjacent_stones:
            self.old_groups.add(adjacent_stone.group)
            self.group.merge(adjacent_stone.group)
        for old_group in self.old_groups:
            self.groups.remove(old_group)
        self.group.set_links_group()
        self.groups.add(self.group)
        return self.group

    def undo(self):
        self.group.remove_stone_from_board(self.stone_link)
        self.groups.remove(self.group)
        for old_group in self.old_groups:
            old_group.set_links_group()
            self.groups.add(old_group)



from os import walk

from ansi_colours import AnsiColours as Colour

from baduk.exception.exceptions import ValidationError
from baduk.game.game import Game
from baduk.service.move_from_sgf import MovesFromSGF
from baduk.ui.action.action import Action
from baduk.ui.pagination import Pagination
from baduk.ui.user_input import UserInput
from baduk.ui.view import View


def move(game, *coordinate):
    try:
        game.move(*coordinate)
        game.board()
    except ValidationError as error:
        game.board()
        View("error.txt").render(error=error)


def rollback(game, number):
    try:
        game.rollback(number)
        game.board()
    except ValidationError as error:
        game.board()
        View("error.txt").render(error=error)


class LoadSGFAction(Action):
    def execute(self, input_value):

        sgfs = []
        for (dir_path, dir_names, file_names) in walk('../sgf'):
            sgfs.extend([[i + 1, file_name] for i, file_name in enumerate(file_names)])
            break
        sgf = Pagination(sgfs, ['key', 'files'])()
        if not sgf:
            return True

        moves = MovesFromSGF('../sgf/' + sgf[1]).get_as_korschelt()
        game = Game(19)
        game.board()
        index = 0
        while index < len(moves):
            user_input = ''
            while user_input not in ['<<', '<', '>', '>>', 'b']:
                user_input = UserInput.get_input('\n%s | %s | %s | %s\n%s Back\n' % (
                    Colour.green("<<"),
                    Colour.green("<"),
                    Colour.green(">"),
                    Colour.green(">>"),
                    Colour.green("b")
                ))
            if user_input == '>':
                move(game, moves[index])
                print("Move: ", moves[index])
                index += 1
                print("Turn: ", index + 1)
                continue
            elif user_input == '>>':
                move(game, *moves[index:index + 10])
                print("Move: ", ", ".join(moves[index:index + 10]))
                index += 10
                print("Turn: ", index + 1)
                continue
            elif user_input == '<':
                index -= 1
                rollback(game, 1)
                print("Turn: ", index + 1)
                continue
            elif user_input == '<<':
                index -= 10
                rollback(game, 10)
                print("Turn: ", index + 1)
                continue
            elif user_input == 'b':
                break
            game.board()
        game.reset()



from baduk.enums.enums import Stone


class StoneLink:
    group = None

    def __init__(self, point, stone):
        self.point = point
        self.stone = stone

    # def __repr__(self):
    #     return '%s: %s' % (str(self.point), str(self.stone))

    def set_group(self, group):
        self.group = group

    def set_stone(self, stone):
        self.stone = stone

    def get_adjacent_stones(self, board):
        return self.point.filter_adjacent_links(
            lambda adjacent_link: self.stone == adjacent_link.stone,
            board
        )

    def get_adjacent_groups(self, board):
        return self.point.filter_adjacent_links(
            lambda adjacent_link: adjacent_link.stone != Stone.NONE,
            board
        )

    def find_adjacent(self, links, liberties, board):
        for adjacent_point in self.point.adjacent_points():
            if board.in_grid(adjacent_point):
                adjacent_link = board.get_point_stone_link(**adjacent_point)
                self.set_links_and_liberties(adjacent_link, board, liberties, links)

    def set_links_and_liberties(self, adjacent_link, board, liberties, links):
        if adjacent_link.stone == Stone.NONE:
            liberties.add(adjacent_link)
        elif adjacent_link.stone == self.stone and adjacent_link not in links:
            links.add(adjacent_link)
            adjacent_link.find_adjacent(links, liberties, board)



class Turn:

    def __init__(self):
        self.turn = 1

    def next_turn(self):
        self.turn += 1

    def get_turn(self):
        return self.turn

    def current_player(self):
        return 1 if self.turn % 2 else 0

    def __str__(self):
        return 'Turn %d' % self.turn

    def reset(self):
        self.turn = 1

    def rollback_turn(self):
        self.turn -= 1



from baduk.command.add_to_adjacent_group_of_stones import AddToAdjacentGroupOfStones
from baduk.command.chain_of_command import ChainOfCommands
from baduk.command.create_group_of_stones import CreateNewGroupOfStones
from baduk.command.merge_with_adjacent_group_of_stones import MergeWithAdjacentGroupsOfStones
from baduk.command.remove_dead_stones import RemoveDeadStones
from baduk.stones.stone_link import StoneLink
from baduk.validation.move_validation import MoveValidation


class GroupOfStonesCollection:

    def __init__(self):
        self.groups = set()
        self.dead_stone_count = 0
        self.chain_of_commands = ChainOfCommands()

    def add_stone_link(self, stone_link: StoneLink, board):
        group = self.add_stone_link_to_group_of_stones(board, stone_link)
        self.update_adjacent_groups_of_stones(board, group, stone_link)
        dead_stones = self.find_dead_stones(group)
        self.dead_stone_count = self.count_dead_stones(dead_stones)
        MoveValidation.check_self_capture(group, self.has_dead_stones(), self.chain_of_commands)
        self.remove_dead_stones(dead_stones)
        MoveValidation.breaks_simple_ko_rule(self, dead_stones, board)

    def update_adjacent_groups_of_stones(self, board, group, stone_link):
        adjacent_groups = {link.group for link in stone_link.get_adjacent_groups(board)}
        adjacent_groups.add(group)
        self.update_groups_of_stones(board, adjacent_groups)

    def update_groups_of_stones(self, board, groups=None):
        for group in groups if groups else self.groups:
            group.update(board)

    def add_stone_link_to_group_of_stones(self, board, stone_link):
        adjacent_stones = list(stone_link.get_adjacent_stones(board))
        if len(adjacent_stones) > 1:
            group = self.chain_of_commands.execute_command(
                MergeWithAdjacentGroupsOfStones(self.groups, stone_link, adjacent_stones)
            )
        elif len(adjacent_stones) == 1:
            group = self.chain_of_commands.execute_command(
                AddToAdjacentGroupOfStones(self.groups, stone_link, adjacent_stones)
            )
        else:
            group = self.chain_of_commands.execute_command(
                CreateNewGroupOfStones(self.groups, stone_link)
            )
        return group

    def find_dead_stones(self, current_group):
        dead_stones = set()
        for group in self.groups:
            # Todo: for class instance comparison classes should implement __hash__ and/or __eq__
            if group.get_liberties() == 0 and group != current_group:
                dead_stones.add(group)
        return dead_stones

    def has_dead_stones(self):
        return self.dead_stone_count > 0

    def count_dead_stones(self, dead_stones):
        count = 0
        for dead_stones in dead_stones:
            count += len(dead_stones.get_links()) if dead_stones else 0
        return count

    def get_dead_stones_count(self):
        return self.dead_stone_count

    def remove_dead_stones(self, dead_stones):
        self.chain_of_commands.execute_command(
            RemoveDeadStones(self.groups, dead_stones)
        )

    def rollback(self):
        self.chain_of_commands.undo()
        self.chain_of_commands.undo()

    def reset(self):
        self.groups = set()
        self.dead_stone_count = 0
        self.chain_of_commands.reset()
